"""
`core_df` contains classes and functions that allow you to manipulate editable dataframes.
"""
import copy
from abc import ABCMeta
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Type, Union

import pyarrow  # type: ignore

from .core import MutableValue, import_failure
from .ss_types import DataFrameRecordAdded, DataFrameRecordRemoved, DataFrameRecordUpdated

if TYPE_CHECKING:
    import pandas
    import polars


class DataframeRecordProcessor():
    """
    This interface defines the signature of the methods to process the events of a
    dataframe compatible with EditableDataframe.

    A Dataframe can be any structure composed of tabular data.

    This class defines the signature of the methods to be implemented.
    """
    __metaclass__ = ABCMeta

    @staticmethod
    def match(df: Any) -> bool:
        """
        This method checks if the dataframe is compatible with the processor.
        """
        raise NotImplementedError

    @staticmethod
    def record(df: Any, record_index: int) -> dict:
        """
        This method read a record at the given line and get it back as dictionary

        >>> edf = EditableDataFrame(df)
        >>> r = edf.record(1)
        """
        raise NotImplementedError

    @staticmethod
    def record_add(df: Any, payload: DataFrameRecordAdded) -> Any:
        """
        signature of the methods to be implemented to process wf-dataframe-add event

        >>> edf = EditableDataFrame(df)
        >>> edf.record_add({"record": {"a": 1, "b": 2}})
        """
        raise NotImplementedError

    @staticmethod
    def record_update(df: Any, payload: DataFrameRecordUpdated) -> Any:
        """
        signature of the methods to be implemented to process wf-dataframe-update event

        >>> edf = EditableDataFrame(df)
        >>> edf.record_update({"record_index": 12, "record": {"a": 1, "b": 2}})
        """
        raise NotImplementedError

    @staticmethod
    def record_remove(df: Any, payload: DataFrameRecordRemoved) -> Any:
        """
        signature of the methods to be implemented to process wf-dataframe-action event

        >>> edf = EditableDataFrame(df)
        >>> edf.record_remove({"record_index": 12})
        """
        raise NotImplementedError

    @staticmethod
    def pyarrow_table(df: Any) -> pyarrow.Table:
        """
        Serializes the dataframe into a pyarrow table
        """
        raise NotImplementedError


class PandasRecordProcessor(DataframeRecordProcessor):
    """
    PandasRecordProcessor processes records from a pandas dataframe saved into an EditableDataframe

    >>> df = pandas.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> edf = EditableDataFrame(df)
    >>> edf.record_add({"a": 1, "b": 2})
    """

    @staticmethod
    @import_failure(rvalue=False)
    def match(df: Any) -> bool:
        import pandas
        return True if isinstance(df, pandas.DataFrame) else False

    @staticmethod
    def record(df: 'pandas.DataFrame', record_index: int) -> dict:
        """

        >>> edf = EditableDataFrame(df)
        >>> r = edf.record(1)
        """
        import pandas

        record = df.iloc[record_index]
        if not isinstance(df.index, pandas.RangeIndex):
            index_list = df.index.tolist()
            record_index_content = index_list[record_index]
            if isinstance(record_index_content, tuple):
                for i, n in enumerate(df.index.names):
                    record[n] = record_index_content[i]
            else:
                record[df.index.names[0]] = record_index_content

        return dict(record)

    @staticmethod
    def record_add(df: 'pandas.DataFrame', payload: DataFrameRecordAdded) -> 'pandas.DataFrame':
        """
        >>> edf = EditableDataFrame(df)
        >>> edf.record_add({"record": {"a": 1, "b": 2}})
        """
        import pandas

        _assert_record_match_pandas_df(df, payload['record'])

        record, index = _split_record_as_pandas_record_and_index(payload['record'], df.index.names)

        if isinstance(df.index, pandas.RangeIndex):
            new_df = pandas.DataFrame([record])
            return pandas.concat([df, new_df], ignore_index=True)
        else:
            new_df = pandas.DataFrame([record], index=[index])
            return pandas.concat([df, new_df])

    @staticmethod
    def record_update(df: 'pandas.DataFrame', payload: DataFrameRecordUpdated) -> 'pandas.DataFrame':
        """
        >>> edf = EditableDataFrame(df)
        >>> edf.record_update({"record_index": 12, "record": {"a": 1, "b": 2}})
        """
        import pandas

        _assert_record_match_pandas_df(df, payload['record'])

        record: dict
        record, index = _split_record_as_pandas_record_and_index(payload['record'], df.index.names)

        record_index = payload['record_index']

        if isinstance(df.index, pandas.RangeIndex):
            df.iloc[record_index] = record  # type: ignore
        else:
            df.iloc[record_index] = record  # type: ignore
            index_list = df.index.tolist()
            index_list[record_index] = index
            df.index = index_list  # type: ignore

        return df

    @staticmethod
    def record_remove(df: 'pandas.DataFrame', payload: DataFrameRecordRemoved) -> 'pandas.DataFrame':
        """
        >>> edf = EditableDataFrame(df)
        >>> edf.record_remove({"record_index": 12})
        """
        record_index: int = payload['record_index']
        idx = df.index[record_index]
        df = df.drop(idx)

        return df

    @staticmethod
    def pyarrow_table(df: 'pandas.DataFrame') -> pyarrow.Table:
        """
        Serializes the dataframe into a pyarrow table
        """
        table = pyarrow.Table.from_pandas(df=df)
        return table


class PolarRecordProcessor(DataframeRecordProcessor):
    """
    PolarRecordProcessor processes records from a polar dataframe saved into an EditableDataframe

    >>> df = polars.DataFrame({"a": [1, 2], "b": [3, 4]})
    >>> edf = EditableDataFrame(df)
    >>> edf.record_add({"record": {"a": 1, "b": 2}})
    """

    @staticmethod
    @import_failure(rvalue=False)
    def match(df: Any) -> bool:
        import polars
        return True if isinstance(df, polars.DataFrame) else False

    @staticmethod
    def record(df: 'polars.DataFrame', record_index: int) -> dict:
        """

        >>> edf = EditableDataFrame(df)
        >>> r = edf.record(1)
        """
        record = {}
        r = df[record_index]
        for c in r.columns:
            record[c] = df[record_index, c]

        return record


    @staticmethod
    def record_add(df: 'polars.DataFrame', payload: DataFrameRecordAdded) -> 'polars.DataFrame':
        _assert_record_match_polar_df(df, payload['record'])

        import polars
        new_df = polars.DataFrame([payload['record']])
        return polars.concat([df, new_df])

    @staticmethod
    def record_update(df: 'polars.DataFrame', payload: DataFrameRecordUpdated) -> 'polars.DataFrame':
        # This implementation works but is not optimal.
        # I didn't find a better way to update a record in polars
        #
        # https://github.com/pola-rs/polars/issues/5973
        _assert_record_match_polar_df(df, payload['record'])

        record = payload['record']
        record_index = payload['record_index']
        for r in record:
            df[record_index, r] = record[r]

        return df

    @staticmethod
    def record_remove(df: 'polars.DataFrame', payload: DataFrameRecordRemoved) -> 'polars.DataFrame':
        import polars

        record_index: int = payload['record_index']
        df_filtered = polars.concat([df[:record_index], df[record_index + 1:]])
        return df_filtered

    @staticmethod
    def pyarrow_table(df: 'polars.DataFrame') -> pyarrow.Table:
        """
        Serializes the dataframe into a pyarrow table
        """
        import pyarrow.interchange  # type: ignore
        table: pyarrow.Table = pyarrow.interchange.from_dataframe(df)
        return table

class RecordListRecordProcessor(DataframeRecordProcessor):
    """
    RecordListRecordProcessor processes records from a list of record saved into an EditableDataframe

    >>> df = [{"a": 1, "b": 2}, {"a": 3, "b": 4}]
    >>> edf = EditableDataFrame(df)
    >>> edf.record_add({"record": {"a": 1, "b": 2}})
    """

    @staticmethod
    def match(df: Any) -> bool:
        return True if isinstance(df, list) else False


    @staticmethod
    def record(df: List[Dict[str, Any]], record_index: int) -> dict:
        """

        >>> edf = EditableDataFrame(df)
        >>> r = edf.record(1)
        """
        r = df[record_index]
        return copy.copy(r)

    @staticmethod
    def record_add(df: List[Dict[str, Any]], payload: DataFrameRecordAdded) -> List[Dict[str, Any]]:
        _assert_record_match_list_of_records(df, payload['record'])
        df.append(payload['record'])
        return df

    @staticmethod
    def record_update(df: List[Dict[str, Any]], payload: DataFrameRecordUpdated) -> List[Dict[str, Any]]:
        _assert_record_match_list_of_records(df, payload['record'])

        record_index = payload['record_index']
        record = payload['record']

        df[record_index] = record
        return df

    @staticmethod
    def record_remove(df: List[Dict[str, Any]], payload: DataFrameRecordRemoved) -> List[Dict[str, Any]]:
        del(df[payload['record_index']])
        return df

    @staticmethod
    def pyarrow_table(df: List[Dict[str, Any]]) -> pyarrow.Table:
        """
        Serializes the dataframe into a pyarrow table
        """
        column_names = list(df[0].keys())
        columns = {key: [record[key] for record in df] for key in column_names}

        pyarrow_columns = {key: pyarrow.array(values) for key, values in columns.items()}
        schema = pyarrow.schema([(key, pyarrow_columns[key].type) for key in pyarrow_columns])
        table = pyarrow.Table.from_arrays(
            [pyarrow_columns[key] for key in column_names],
            schema=schema
        )

        return table

class EditableDataFrame(MutableValue):
    """
    Editable Dataframe makes it easier to process events from components
    that modify a dataframe like the dataframe editor.

    >>> initial_state = wf.init_state({
    >>>    "df": wf.EditableDataFrame(df)
    >>> })

    Editable Dataframe is compatible with a pandas, thrillers or record list dataframe
    """
    processors = [PandasRecordProcessor, PolarRecordProcessor, RecordListRecordProcessor]

    def __init__(self, df: Union['pandas.DataFrame', 'polars.DataFrame', List[dict]]):
        super().__init__()
        self._df = df
        self.processor: Type[DataframeRecordProcessor]
        for processor in self.processors:
            if processor.match(self.df):
                self.processor = processor
                break

        if self.processor is None:
            raise ValueError("The dataframe must be a pandas, polar Dataframe or a list of record")

    @property
    def df(self) -> Union['pandas.DataFrame', 'polars.DataFrame', List[dict]]:
        return self._df

    @df.setter
    def df(self, value: Union['pandas.DataFrame', 'polars.DataFrame', List[dict]]) -> None:
        self._df = value
        self.mutate()

    def record_add(self, payload: DataFrameRecordAdded) -> None:
        """
        Adds a record to the dataframe

        >>> df = pandas.DataFrame({"a": [1, 2], "b": [3, 4]})
        >>> edf = EditableDataFrame(df)
        >>> edf.record_add({"record": {"a": 1, "b": 2}})
        """
        assert self.processor is not None

        self._df = self.processor.record_add(self.df, payload)
        self.mutate()

    def record_update(self, payload: DataFrameRecordUpdated) -> None:
        """
        Updates a record in the dataframe

        The record must be complete otherwise an error is raised (ValueError).
        It must a value for each index / column.

        >>> df = pandas.DataFrame({"a": [1, 2], "b": [3, 4]})
        >>> edf = EditableDataFrame(df)
        >>> edf.record_update({"record_index": 0, "record": {"a": 2, "b": 2}})
        """
        assert self.processor is not None

        self._df = self.processor.record_update(self.df, payload)
        self.mutate()

    def record_remove(self, payload: DataFrameRecordRemoved) -> None:
        """
        Removes a record from the dataframe

        >>> df = pandas.DataFrame({"a": [1, 2], "b": [3, 4]})
        >>> edf = EditableDataFrame(df)
        >>> edf.record_remove({"record_index": 0})
        """
        assert self.processor is not None

        self._df = self.processor.record_remove(self.df, payload)
        self.mutate()

    def pyarrow_table(self) -> pyarrow.Table:
        """
        Serializes the dataframe into a pyarrow table

        This mechanism is used for serializing data for transmission to the frontend.

        >>> df = pandas.DataFrame({"a": [1, 2], "b": [3, 4]})
        >>> edf = EditableDataFrame(df)
        >>> pa_table = edf.pyarrow_table()
        """
        assert self.processor is not None

        pa_table = self.processor.pyarrow_table(self.df)
        return pa_table

    def record(self, record_index: int):
        """
        Retrieves a specific record in dictionary form.

        :param record_index:
        :return:
        """
        assert self.processor is not None

        record = self.processor.record(self.df, record_index)
        return record



def _assert_record_match_pandas_df(df: 'pandas.DataFrame', record: Dict[str, Any]) -> None:
    """
    Asserts that the record matches the dataframe columns & index

    >>> _assert_record_match_pandas_df(pandas.DataFrame({"a": [1, 2], "b": [3, 4]}), {"a": 1, "b": 2})
    """
    import pandas

    columns = set(list(df.columns.values) + df.index.names) if isinstance(df.index, pandas.RangeIndex) is False else set(df.columns.values)
    columns_record = set(record.keys())
    if columns != columns_record:
        raise ValueError(f"Columns mismatch. Expected {columns}, got {columns_record}")

def _assert_record_match_polar_df(df: 'polars.DataFrame', record: Dict[str, Any]) -> None:
    """
    Asserts that the record matches the columns of polar dataframe

    >>> _assert_record_match_pandas_df(polars.DataFrame({"a": [1, 2], "b": [3, 4]}), {"a": 1, "b": 2})
    """
    columns = set(df.columns)
    columns_record = set(record.keys())
    if columns != columns_record:
        raise ValueError(f"Columns mismatch. Expected {columns}, got {columns_record}")

def _assert_record_match_list_of_records(df: List[Dict[str, Any]], record: Dict[str, Any]) -> None:
    """
    Asserts that the record matches the key in the record list (it use the first record to check)

    >>> _assert_record_match_list_of_records([{"a": 1, "b": 2}, {"a": 3, "b": 4}], {"a": 1, "b": 2})
    """
    if len(df) == 0:
        return

    columns = set(df[0].keys())
    columns_record = set(record.keys())
    if columns != columns_record:
        raise ValueError(f"Columns mismatch. Expected {columns}, got {columns_record}")



def _split_record_as_pandas_record_and_index(param: dict, index_columns: list) -> Tuple[dict, tuple]:
    """
    Separates a record into the record part and the index part to be able to
    create or update a row in a dataframe.

    >>> record, index = _split_record_as_pandas_record_and_index({"a": 1, "b": 2}, ["a"])
    >>> print(record) # {"b": 2}
    >>> print(index) # (1,)
    """
    final_record = {}
    final_index = []
    for key, value in param.items():
        if key in index_columns:
            final_index.append(value)
        else:
            final_record[key] = value

    return final_record, tuple(final_index)



import re

import writer as wf
import writer.ai

# Welcome to Writer Framework! 
# This is a simple app to get you started with text completion.
# More documentation is available at https://dev.writer.com

def handle_button_click(state):
    state["message"] = "% Loading up expert social media posts..."
    
    prompt = f"You are a social media expert. Generate 5 engaging social media posts about {state['topic']}. Include emojis."
    state["posts"] = writer.ai.complete(prompt)

    prompt = f"You are a social media expert. Generate 5 hashtags about {state['topic']}, delimited by spaces. For example, #dogs #cats #ducks #elephants #badgers"
    pattern = r'#\w+'
    hashtags = re.findall(pattern, writer.ai.complete(prompt))
    state["tags"] = {item: item for item in hashtags}
    
    state["message"] = ""

# Initialize state here

wf.init_state({
    "posts": "",
    "topic": "writing",
    "message": ""
})


import writer as wf

# This is a placeholder to get you started or refresh your memory.
# Delete it or adapt it as necessary.
# Documentation is available at https://dev.writer.com/framework

# Shows in the log when the app starts
print("Hello world!")

# Its name starts with _, so this function won't be exposed
def _update_message(state):
    is_even = state["counter"] % 2 == 0
    message = ("+Even" if is_even else "-Odd")
    state["message"] = message

def decrement(state):
    state["counter"] -= 1
    _update_message(state)

def increment(state):
    state["counter"] += 1
    # Shows in the log when the event handler is run
    print("The counter has been incremented.")
    _update_message(state)
    
# Initialize the state

# "_my_private_element" won't be serialized or sent to the frontend,
# because it starts with an underscore

initial_state = wf.init_state({
    "my_app": {
        "title": "My app"
    },
    "_my_private_element": 1337,
    "message": None,
    "counter": 26,
})

_update_message(initial_state)



import hashlib
import os

from fastapi import HTTPException, Request

HASH_SALT = "a9zHYfIeL0"

def get_hash(message: str):
    base_hash = os.getenv("WRITER_SECRET_KEY")
    if not base_hash:
        raise ValueError("Environment variable WRITER_SECRET_KEY needs to be set up in" + \
                            "order to enable operations which require hash generation, such as creating async jobs.")
    assert HASH_SALT
    combined = base_hash + HASH_SALT + message
    return hashlib.sha256(combined.encode()).hexdigest()

def verify_message_authorization_signature(message: str, request: Request):
    auth_header = request.headers.get("Authorization")
    if not auth_header:
        raise HTTPException(status_code=401, detail="Unauthorized. Token not specified.")
    if auth_header != f"Bearer {get_hash(message)}":
        raise HTTPException(status_code=403, detail="Forbidden. Incorrect token.")


import importlib.metadata
import logging
import textwrap
from types import ModuleType
from typing import Any, Dict, List, Optional, Type, TypeVar, Union, cast

from writer.core import (
    BytesWrapper,
    Config,
    FileWrapper,
    Readable,
    State,
    WriterState,
    base_component_tree,
    get_app_process,
    initial_state,
    new_initial_state,
    session_manager,
    session_verifier,
)
from writer.core import (
    writerproperty as property,
)
from writer.core_df import EditableDataFrame

try:
    from writer.ui import WriterUIManager
    PROPER_UI_INIT = True
except ModuleNotFoundError:
    logging.error(
            textwrap.dedent(
                """\
                \x1b[31;20mError: Failed to import `writer.ui` module.
                This indicates that the Writer Framework was not built properly.
                Please refer to CONTRIBUTING.md for instructions to resolve this issue.\x1b[0m"""
            )
        )
    PROPER_UI_INIT = False

def _get_ui_runtime_error_message() -> str:
    return "UI module is unavailable – Writer Framework is not properly built. " + \
            "Please refer to CONTRIBUTING.md for instructions to resolve this issue."

VERSION = importlib.metadata.version("writer")

base_component_tree
session_manager
Config
session_verifier


def pack_file(file: Union[Readable, str], mime_type: Optional[str] = None):
    """
    Returns a FileWrapper for the file provided, which is automatically
    serialised to a data URL.
    """

    return FileWrapper(file, mime_type)


def pack_bytes(raw_data, mime_type: Optional[str] = None):
    """
    Returns a BytesWrapper for the bytes raw data provided, which is automatically
    serialised to a data URL.
    """

    return BytesWrapper(raw_data, mime_type)


S = TypeVar('S', bound=WriterState)


def init_ui() -> 'WriterUIManager':
    """Initializes and returns an instance of WriterUIManager.
    This manager provides methods to dynamically create and manage UI
    components in a Writer Framework application.

    The WriterUIManager allows for the creation of application-wide,
    code-managed components during startup, ensuring that the set of components
    is initially accessible by all sessions.

    :return: An instance that serves as a bridge for programmatically
    interacting with the frontend, facilitating dynamic UI component management.
    :rtype: WriterUIManager

    **Example**::

    >>> import writer as wf
    >>>
    >>> with wf.init_ui() as ui:
    >>>     with ui.Page({"key": "hello"}):
    >>>         ui.Text({"text": "Hello pigeons"})
    """
    if PROPER_UI_INIT:
        return WriterUIManager()
    else:
        raise RuntimeError(_get_ui_runtime_error_message())


def init_state(raw_state: Dict[str, Any], schema: Optional[Type[S]] = None) -> Union[S, WriterState]:
    """
    Sets the initial state, which will be used as the starting point for
    every session.

    initial_state.user_state.state = {}
    initial_state.user_state.ingest(state_dict)
    return initial_state



    >>> import writer as wf

    >>> initial_state = wf.init_state({
    >>>   "counter": 0,
    >>> }, schema=AppSchema)
    """
    concrete_schema = cast(Type[S], WriterState if schema is None else schema)
    if not issubclass(concrete_schema, WriterState):
        raise ValueError("Root schema must inherit from WriterState")

    _initial_state: S = new_initial_state(concrete_schema, raw_state)

    return _initial_state


def init_handlers(handler_modules: Union[List[ModuleType], ModuleType]):
    """
    Registers one or more handler modules to enable its containing functions
    to be used as event handlers by the application's frontend.

    :param handler_modules: A module or list of modules for registration.
    :type handler_modules: module or list of modules

    **Examples**

    Register a single handler module:

    >>> import writer as wf
    >>> import my_handler_module
    >>> wf.init_handlers(my_handler_module)

    Register multiple handler modules:

    >>> import writer as wf
    >>> import module_one, module_two
    >>> wf.init_handlers([module_one, module_two])

    :raises ValueError: If an object that is not a module is attempted to be registered.
    """
    from writer.core import get_app_process
    current_app_process = get_app_process()
    handler_registry = current_app_process.handler_registry
    # Ensure handler_modules is a list
    if not isinstance(handler_modules, list):
        handler_modules = [handler_modules]

    for module in handler_modules:
        handler_registry.register_module(module)


def middleware():
    """
    A "middleware" is a function that works with every event handler before it is processed and also before returning it.

    >>> import writer as wf
    >>>
    >>> @wf.middleware()
    >>> def my_middleware(state):
    >>>     state['processing'] += 1
    >>>     yield
    >>>     state['processing'] -= 1

    Middleware accepts the same arguments as an event handler.

    >>> import writer as wf
    >>>
    >>> @wf.middleware()
    >>> def my_middleware(state, payload, session):
    >>>     state['processing'] += 1
    >>>     yield
    >>>     state['processing'] -= 1
    """
    def inner(func):
        _app_process = get_app_process()
        _app_process.middleware_registry.register(func)


    return inner
